// 																																			KEDS Project
// 																												(Kansas Event Data System)
//__________________________________________________________________________________
// 											          Count_Conflict: windowed tabulation of WEIS events 

// This program goes through and counts the number of events above WEIS_THRESHOLD
// in the DATE_WINDOW period following each week.  It tabulates all of the dyadic
// behavior among the groups listed in ACTOR_LIST.

// Output: Text file "Count_Confl.out" contains the compilation.  There is no 
//         feedback while the program is running.

//	Programmer: Philip A. Schrodt
//	Compiler:		Metrowerks C
//	Language:		ANSI C except for // comments

//	Revision History: 
//		99.11		Original program -- JWAC Central Asia project
//		00.6		Revised for West Africa analysis; compiler directives standardized

//__________________________________________________________________________________

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

//__________________________________________________________________________________
// 	Macros

#define	FALSE			0
#define TRUE			1

#define IN_FILE		"WAFR.events.rand"
//#define IN_FILE		"WAFR.events"
//#define IN_FILE		"AFG.ALL.EVT"
//#define IN_FILE		"CASIA.ALL.EVT.3CODE"
//#define IN_FILE		"CASIA.LEADS.EVT.3CODE"
#define OUT_FILE	"Count_Confl.out"

//#define ACTOR_LIST "ARM.AZR."  // codes for the dyads we are interested in,
//#define ACTOR_LIST "RUS.TAJ."  // separated by . to avoid accidental matches 
//#define ACTOR_LIST "AFG.AFO.RUS."
#define ACTOR_LIST "LBR.NIG.ECO." 
//#define ACTOR_LIST "SIE.NIG.ECO."
//#define ACTOR_LIST "NIG." 
//#define ACTOR_LIST "NIG.CAO." 

//#define	DATE_WINDOW  28      // width of date window in days
#define	DATE_WINDOW  56 

// Julian bounds for the entire data set (used to avoid invalid dates)

// Central Asia analysis
//#define	HIGH_DAY		34910		//	99.07.31
//#define	LOW_DAY   	31168		//	89.05.02

// West Africa analysis
#define	HIGH_DAY		35063		//	99.12.31
#define	LOW_DAY   	31050		//	89.01.04

// Julian bounds for a specific dyad: Central Asia analysis

// Central Asia analysis
//#define	LAST_DAY		34910		//	99.07.31
//#define	FIRST_DAY   32142		//	92.01.01

// West Africa analysis
#define	LAST_DAY		35063		//	99.12.31
#define	FIRST_DAY   31050		//	89.01.04


#define N_POINTS    	 40000      // maximum number of data points 
#define N_DISTN     			11			// gives the number of bins (width 2) in the distribution summary
#define WEIS_THRESHOLD	 200			// 3-digit WEIS code for threshold of "violent" events

//___________________________________________________________________________
// Prototypes

long Julian (char *s);	// Converts dates of form YYMMDD to Julian serial beginning 0 = 1 Jan 1904
void Read_Data(void);   // read the entire data set into arrays

//__________________________________________________________
// Globals


int		Month[] = {0,0,31,59,90,120,151,181,212,243,273,304,334};		// used for Julian computation: # days prior to Month[n] JAN=1
long  adate[N_POINTS];
char  asrc[N_POINTS][7];
char  atar[N_POINTS][7];
int   acode[N_POINTS];
long	npoints = 0;
char  sinterest[]= ACTOR_LIST;  // codes for the dyads we are interested in, separated by . to avid accidental matches 
int		distn[N_DISTN];								// distribution of values
  
//__________________________________________________________________________________
//																																			main program

void main(void)
{
FILE	*fout;				// output file
int 	nweek = FIRST_DAY;
int		ijul;
int   ka = 0;
int   kviol=0,knvio=0;				// count of violent, nonviolent events
int   start_week = 0;			// starting point of week in array
int   kweeks = 0;

	Read_Data();
	fout = fopen(OUT_FILE,"w");
	if (!fout) { printf("Error: can't open %s",OUT_FILE); exit(0); }
	fprintf(fout,"Input file: %s\n",IN_FILE);
	fprintf(fout,"Interest codes: %s\n",sinterest);
	fprintf(fout,"Window: %d days\n",DATE_WINDOW);
	fprintf(fout,"\nDate\tWEIS>=%3d\tWEIS<%3d\n", WEIS_THRESHOLD,WEIS_THRESHOLD);
	
	for (ka=0; ka<N_DISTN; ka++) distn[ka]=0;   // supposed to zero already but make sure
	
	 while (nweek<LAST_DAY - DATE_WINDOW)
	{
		while (adate[start_week] < nweek) start_week++;  // find the first data point in the array
		ka = start_week;
		ijul = adate[ka];
		while (ijul < nweek + DATE_WINDOW)
		{
			if (strstr(sinterest,asrc[ka]) && strstr(sinterest,atar[ka]))
			{	if (acode[ka] >= WEIS_THRESHOLD) 
							kviol++;
				else	knvio++; 
			}
			ka++;
			if (ka>=npoints) break;
			ijul = adate[ka];
		}
		fprintf(fout,"%ld\t%4d\t%4d\n",nweek,kviol,knvio);

		kviol /=2;					// record the distribution; bin size is 2
		if (kviol >= N_DISTN) kviol= N_DISTN-1;
		distn[kviol]++;
		kweeks++;
		
		kviol  = 0;
		knvio  = 0;
		nweek += 7;
	}
	
	fprintf(fout,"\nDistribution\n");
	for (ka=0; ka<N_DISTN-1; ka++) fprintf(fout,"%02d-%02d\t%4d\t%6.1f%%\n",ka*2,1+ka*2,distn[ka],100*(float)distn[ka]/kweeks);
	ka = N_DISTN-1;		// again, it probably is this anyway...
	fprintf(fout,">=%02d \t%4d\t%6.1f%%\n",ka*2,distn[ka],100*(float)distn[ka]/kweeks);

	fclose(fout);
	printf("Finished.");
	
}

long Julian (char *s)
// Converts dates of form YYMMDD to Julian serial beginning 0 = 1 Jan 1904
// (which matches the Excel convention) 
{
int 	mo,dy;
long	yr, ij;

  for (mo=0; mo<7; mo++) if (s[mo] == ' ') s[mo] = '0';			// convert blanks to zeros
  
																		//Break date out of the string and convert to integers 
	yr = (short)s[0] * 10 + (short)s[1] - 532;			// 532 substracts the 1904 base from yr
	mo = (short)s[2] * 10 + (short)s[3] - 528;
	dy = (short)s[4] * 10 + (short)s[5] - 528;
						
	ij =  yr * 365;
	ij += ((yr+3) / 4) + Month[mo] + (dy-1);				// ((yr+3)/4) is the leap-year adjustment
	if (((yr % 4) == 0) && (mo > 2)) ij++;	// adjust for month within a leap-year
	return ij;
} // Julian

void Read_Data(void)
// read the entire data set into arrays
{
FILE	*fin;				// input and output files
char 	stail[40];	// read junk from end of line
char  date[7];		// character version of date
int 	ijul;				// julian value of date
int		ka = 0;			// array index

	fin = fopen(IN_FILE,"r");
	if (!fin) { printf("Error: can't open %s",IN_FILE); exit(0); }
	
	while (!feof(fin))
	{
		fscanf(fin,"%s %s %s %d",&date, asrc[ka], atar[ka], &acode[ka]);
		fgets(stail,40,fin);
		asrc[ka][3] = '\0';		// truncate to 3-char codes
		atar[ka][3] = '\0';
		ijul = Julian(date);
		if ((ijul<LOW_DAY) || (ijul>HIGH_DAY))					// check this because we don't trust Reuters
			printf("Date error:%s\n",date);
		else
		{ adate[ka] = Julian(date);
		  ka++;
		}
	}
	
	npoints = ka;		// transfer value to the global
	fclose(fin);	// we're finished with this
} // Read_Data