// 																														JWAC Forecasting Project
// 																															(Central Asia models)
//____________________________________________________________________________________
// 																		  													CASIASQ/WEISSQ Filter

//  This program converts WEIS event files to a single multiple-dyad sequence that 
//  can be analyzed using a hidden Markov model.  It uses 2-digit WEIS codes; the
//  codes for the n-th dyad are 
//			WEIS + (n-1)*22
//  Days where no event occurs are filled with a single '00' code
//  Events are put into chronological order with a Julian date base 1904.1.1 . 

//	INPUT FILES
//		DATA_FILE			KEDS event files
//
//		DYAD_FILE			List of dyads to compile, *** indicates "all actors"
//									First line contains file Suffix
//									File is terminated with =======
//									
//

//	OUTPUT FILES
//		WEISSQ.<suffix>		Output of a directed dyad in the format: <Julian date> <2-digit WEIS code>
//		DIST.<suffix>			Distribution of the event categories
//
//		An assortment of additional printing options have been commented-out.

//	Notes:
//
//	1. Make_Random() will generate a random sequence with a known marginal distribution
//
//	2. If the DO_XFORM option is set, the program will re-code the 2-digit codes.
//     The recoding is set in 'Initialize'; the constant N_XFORM_CODES also needs
//     to be set.     

//	Programmer: Philip A. Schrodt
//	Compiler:		Metrowerks C
//	Language:		ANSI C except for "//" comments

//	Revision History: 
//		97.2		Original KEDS project program
//    97.8		Modified to work on APSA97 file; switched to the Excel 1904 Julian base;
//						Added dyadic output option and Make_Random().
//		99.2    Minimally modified to work on LEV79-99.evt file
//		99.3		Modified to generate multiple dyads for JWAC project
//    99.9    Added the DO_XFORM option to handle recoding
//    99.11   Modified from WEISSQ to do Central Asia recodes

//	Original program development funded by National Science Foundation Grant SBR-9410023. 

//__________________________________________________________________________________

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "CASIASQ_Filter.h"


void main(void)
{
//	Make_Random( );			// generate a random sequence

	Initialize();
	Make_Sequence();	
	Write_Dist();

} // main

void Initialize( void )	
// Open various files
{
char fname[33];
													
																// read the dyad codes
	fin=fopen(DYAD_FILE,"r");
	if (!fin) { printf("Errror: can't open the dyad file %s\n",DYAD_FILE); exit(0); }
	fscanf(fin,"%s",Suffix);
	fscanf(fin,"%s %s",&Src_Code[N_Dyads],&Tar_Code[N_Dyads]); 
	while((!feof(fin)) && strncmp(Src_Code[N_Dyads],DYAD_END_STRG,3))
	{
		if (++N_Dyads >=MAX_DYADS) break;
		fscanf(fin,"%s %s",&Src_Code[N_Dyads],&Tar_Code[N_Dyads]); 
	}
	fclose(fin);
		
																// open the event data file
	fin=fopen(DATA_FILE,"r");
	if (!fin) { printf("Errror: can't open the event data file %s\n",DATA_FILE); exit(0); }
	
																// open the output file
	strcpy(fname,OUT_PREFIX);
	strcat(fname,Suffix);
	fout=fopen(fname,"w");
	if (!fout) { printf("Errror: can't open output file %s\n",fname); exit(0); }
	
#if DO_XFORM
	// initialize the transform file (this is done using this table to reduce the 
	// likelihood of miscoding...)

	Xform_Codes[0]  = 0;
	Xform_Codes[1]  = 2;
	Xform_Codes[2]  = 1;
	Xform_Codes[3]  = 1;
	Xform_Codes[4]  = 1;
	Xform_Codes[5]  = 1;
	Xform_Codes[6]  = 2;
	Xform_Codes[7]  = 2;
	Xform_Codes[8]  = 1;
	Xform_Codes[9]  = 1;
	Xform_Codes[10] = 1;
	Xform_Codes[11] = 3;
	Xform_Codes[12] = 3;
	Xform_Codes[13] = 3;
	Xform_Codes[14] = 3;
	Xform_Codes[15] = 3;
	Xform_Codes[16] = 3;
	Xform_Codes[17] = 3;
	Xform_Codes[18] = 4;
	Xform_Codes[19] = 4;
	Xform_Codes[20] = 4;
	Xform_Codes[21] = 4;
	Xform_Codes[22] = 4;
#endif	

} // Initialize


void Make_Sequence(void)
// generate the sequence
{
long		julLast;																// julian value of last date written plus 1
long		julCurrent;															// julian value of currnet date
char		sdate[DATE_LENGTH];											// date code
char		sscr[CODE_LENGTH], star[CODE_LENGTH];		// source and target codes
int			ievt;																		// event code
char		dataline[LINELENGTH];										// tail of the input string
int			ka;

	fscanf(fin,"%s",sdate);
	julLast = Julian(sdate)+1;
	rewind(fin);
	while (!feof(fin))
	{
		fscanf(fin,"%s %s %s %d",sdate,sscr,star,&ievt);	// read the data
		fgets(dataline,LINELENGTH,fin);									// read the rest of the line
		ievt /=10;																			// reduce to 2-digit code

#if DO_XFORM			// transform the codes
		if ((ievt>=0) && (ievt<=N_EVT_CODES)) ievt = Xform_Codes[ievt];
		else
		{
		  printf("Bad code: %s  %d\n",sdate,ievt);
		  ievt = 0;
		 }
#endif
																	
		for (ka=0; ka<N_Dyads; ka++)										// see if we've got one of the dyads
		{
			if (!strcmp(WILD_CARD_CODE,Src_Code[ka]) || !strcmp(Src_Code[ka],sscr))
			{
				if (!strcmp(WILD_CARD_CODE,Tar_Code[ka]) || !strcmp(Tar_Code[ka],star))
				{	
			 		julCurrent = Julian(sdate);			// we've got one, so convert the date

					if ( julLast < julCurrent ) 		// need to write intermediate dates
					{
						while ( julLast < julCurrent ) 							// fill in null events on intermediate days
						{ 	
							fprintf(fout,"%6d      00\n",julLast);	// these are deliberately spaced to the right
							Zero_Dist++;
							julLast++;
						}
						printf("%s\n",sdate);					// use this opportunity to show the program is alive
					}
					julLast = julCurrent+1; 				// reset julLast to the day after 
					
#if DO_XFORM
					fprintf(fout,"%6d  %3d\n",julCurrent, ievt+(ka*N_XFORM_CODES));  // write the event code plus the offset 
#else
					fprintf(fout,"%6d  %3d\n",julCurrent, ievt+(ka*N_EVT_CODES));
#endif
					Evt_Dist[ka][ievt-1]++;						
				}
			}
		}
	}
	
	fclose(fin);
	fclose(fout);

} // Make_Sequence

void Write_Dist(void)
// write the event distribution
// Note: the output format is specific to WEIS
{
int 	ka,kb;
char fname[33];
	
	strcpy(fname,DIST_PREFIX);
	strcat(fname,Suffix);
	fout=fopen(fname,"w");
	if (!fout) { printf("Errror: can't open output file %s\n",fname); exit(0); }
	printf("Writing distribution to %s\n",fname);

	fprintf(fout,"Event Distribution for %s file\n",DYAD_FILE);
	fprintf(fout,"\n00 events:	%5d\n",Zero_Dist);

	for (ka=0;ka<N_Dyads; ka++)
	{
		fprintf(fout,"\nDyad: %s > %s\n\n",Src_Code[ka],Tar_Code[ka]);
		for (kb=0; kb<10; kb++)
			fprintf(fout,"%02d  %5d      %02d  %5d\n",kb+1,Evt_Dist[ka][kb], kb+11,Evt_Dist[ka][kb+10]);
		fprintf(fout,"               21  %5d\n",Evt_Dist[ka][20]);
		fprintf(fout,"               22  %5d\n",Evt_Dist[ka][21]);
	}
	fclose(fout);
	printf("finished...use Cmd-Q to exit");
} // Write_Dist

#if DO_RANDOM

double UnifRnd( void )
// Uniform[0,1] random number generator
{
return  ((double)rand())/RAND_MAX;
}

int Get_Random_Event ( void )
{
int ka;
float ran;

	ran = UnifRnd();
	ka=0;
	while (ran > evtprob[ka])
	 ka++;
	return ka;
} // Get_Random_Event

void Make_Random( void )
// generate a random sequence of events based on the marginals in evtprob[]
{
#define LASTJUL			 34129
long		julCurrent = 27498;		// Julian date for 790415
int			ka,kb;
float		ran;
																			
	for (ka=1; ka<22; ka++) evtprob[ka] += evtprob[ka-1];		// convert evtprob to cumulative probs
	evtprob[22]=1.0;
	
	printf("\nGenerating random sequence...\n");
	fout=fopen("WEISDY Random","w");
	if (!fout) { printf("Errror: can't open WEISDY Random"); exit(0); }

//	fprintf(fout,"\n%3.3s > %3.3s",ActorCode,TargetCode);
//	printf("======================\n");
	while (julCurrent < LASTJUL)
	{
		ka=Get_Random_Event();
		if ((ka) && (UnifRnd() < 0.33)) ka += 22;			// shift to an obverse event
		fprintf(fout,"%6d  %2d\n",julCurrent, ka);
		if (ka)
		{ while (UnifRnd() < 0.2)     // possible repeat events on nonzero days
			{ do
				{
					ka=Get_Random_Event();	// get a nonzero event
					if ((ka) && (UnifRnd() < 0.33)) ka += 22;			// shift to an obverse event
				}
				while (!(ka));
			fprintf(fout,"%6d  %2d\n",julCurrent, ka);	
			}
			julCurrent++; 
		}
	}
	fclose(fout);
	printf("finished...");
} // Make_Random

#endif	// DO_RANDOM

long Julian (char *s)
// Converts dates of form YYMMDD to Julian serial beginning 0 = 1 Jan 1904
// (which matches the Excel convention) 
{
int 	mo,dy;
long	yr, ij;

  for (mo=0; mo<7; mo++) if (s[mo] == ' ') s[mo] = '0';			// convert blanks to zeros
  if (s[0] == '0') s[0] = ':';															// increment the 00 to 09 dates
  
																		//Break date out of the string and convert to integers 
	yr = (short)s[0] * 10 + (short)s[1] - 532;			// 532 substracts the 1904 base from yr
	mo = (short)s[2] * 10 + (short)s[3] - 528;
	dy = (short)s[4] * 10 + (short)s[5] - 528;
						
	ij =  yr * 365;
	ij += ((yr+3) / 4) + Month[mo] + (dy-1);				// ((yr+3)/4) is the leap-year adjustment
	if (((yr % 4) == 0) && (mo > 2)) ij++;	// adjust for month within a leap-year
	return ij;
} // Julian
