// 																																			KEDS Project
// 																												(Kansas Event Data System)
//__________________________________________________________________________________
//	Left-Right-Left Hidden Markov model for international event sequences 	HMM.LRL 

// CODE MODIFIED FOR PURE PREDICTION 99.11.27
// This version does purely predictive modelling, stepping through each year and 
// getting templates only from the previous period.

// <99.6.28> Additional modification to Compare_HighLow and Compute_WeeklyProb so
//           this compute the accuracy only on the prior weeks.  The complete 
//           forecast is still written to the file; the accuracy calculations
//           are taken care of in the pr.harvest.c program

// MODIFICATIONS FOR CENTRAL ASIA TESTS: 99.11.8  (HMM.CASIA.FS)
// 1. Modified the XFRM variant to handle the three CASIA cases.  Also removed 
//    the old cross-correlation and TFT code so we don't have to scroll through
//    it.  This is the full-sample test.
//
// 2. Note the addition of the BASE_DATE constant to set the first day of
//    data that will be used in the analysis.  The .SQ files contain data
//    going back to 1989, but the analysis goes from 1992 forward.

// MODIFICATIONS FOR NCSA GA NT/XFRM TESTS: 99.9.8
// 1. Un-did the short-circuiting of the 'Evolve' procedure.  This is now being used
//    to test the effects of increasing the number of templates, and the reduced-code
//    model.

// MODIFICATIONS FOR NCSA MONTE-CARLO TEST: 99.6.9
// 1. This is a minor modification of hmm.ga.sc.c to do a comparative test of
//    Monte-Carlo estimation versus the GA estimation.  The change is simply to
//    short-circuit the reproduction in 'Evolve', replacing this with a simple
//    random parameter setting; everything else stays the same.
//
// 2. MPI CODE (99.7.10) -- Added the minimal amount of MPI code to get this thing
//    to run in parallel in the outermost loop, saving separate files that can be
//    re-combined later.


// CODE AS SAVED 99.3.16.0900
// This estimates a high model using a GA, then [independently] estimates a low model, 
// then compares the two on a week-by-week basis, then writes it all to a file
// Output results: GA.BALKANS.3.16.1

// MODIFICATIONS: 3.16
// 1. The size of the GA_VECT arrays has been doubled, so we are now dealing with
//    pairs of vectors, one for the high estimate and one for the low estimate.

//	 This program is designed to estimate and fit WEIS event sequences using a 
//	left-right-left Hidden Markov model.

//	Modifications by Philip A. Schrodt (p-schrodt@ukans.edu)
//	Date: 97.3.10.
	  
//	1. Program is based on C++ code originally written by Richard Myers 
//	   and James Whitson;see the header file for detailed copyright and
//	   modification history.

//	2. Files were translated into Metrowerks C (Macintosh).  This involved moving
//	   from the object model of C++ to a bunch of clunky globals and combining
//	   the various original programs.  Less elegant but it works...

//	3. Modified the input scheme to work with BCOW sequences -- which are rather
//		 long -- and WEIS codes.  

//	4. The functions under "continuous WEIS sequences" do a running computation of
//		 the alphas for a given .hmm matrix by month, doing a fit of the previous 
//		 100 days in each case.  This produces a tab-delimited file. 
 
//	COMMENTS ON CURRENT CODE <97.8.12>
//	This thing got built quite quickly in order to get estimates for the APSA 
//	paper, and there is a *lot* of idiosyncratic control via #if statements and
//	commenting-out.  It all makes sense given the tasks in the paper but at the
//	moment is this *not a general program* -- instead it is full of idiosyncracies
//	that perform specific tasks required for the analysis.
//	
//	At the moment these include:
//	1. Program generally writes over files; the assumption is that the names of
//	   earlier versions of the files will have been changed if they need to be
//	   saved.
//	   
//	2. Some of the routines do not include the 1982-1985 period; others do.
//	
//	3. qprob at the moment looks at two consecutive states, not just one.
//	
//	4. The file-structure documentation below is generally out of date.
//	
//	5. The cross-correlation function XCORR has a quirky if() that gets around
//	   a splice in the data set.
//	
    
//____________________     
//	HMM MODELS
//		Line 1		Number of states
//		Line 2		Number of symbols
//		remainder	pp	pr	pn	s1  s2  ...  sn
//							pp = prob of transition to previous state
//							pr = prob of recurrence 
//							pp = prob of transition to next state 
//							s1, s2,...,sn = symbol probabilities 

//____________________
//	WEIS SEQUENCES
//	Input data file: 
//		WEISME SSS>TTT	Sequence of WEIS events, one per line.  
//										SSS = sSource[]; TTT = sTarget[]
//
//		initial_model		Follows the Myers and Whitson format; usually this is the
//										produced by the Train_HMM() function. 		


//	Output: 
//		WEIS.SSS>TTT.alphas  	Alphas by month, tab-delimited with YYMMDD dates in 
//													first field

//			stdout							same as above

//____________________
//	OPTIONS:
//		At present, control of the functions of the program are managed by #ifs and 
//		commenting-out in main().

//____________________
//
//	Required files:
//	hmm.ga.h					header
//	hmm.estimate.c		estimation routines and utilities
//	hmm.ga.c					main and genetic algorithm


//	Programmer: Philip A. Schrodt
//	Compiler:		Metrowerks C
//	Language:		ANSI C except for // comments

//	Revision History: 
//		97.2		Original modification of Meyers & Whitson program
//	  97.6		Modification to handle random starting points; Viterbi algorithm
//	  97.8		Modification to estimate various LRL models; date-based templates;
//	    			cross-correlation optimization
//	  99.3		Modification to do GA and Balkans model estimation
//	  99.11		Modification to do Central Asia model estimation
	    
//__________________________________________________________________________________

#include <assert.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <mpi.h>
#include "hmm.wafrc.pr.h"

//__________________________________________________________________________________
//																																			main program

void main(int *argc, char** argv)
{

int			nex;
time_t tStart,tEnd;
	
int size, rank;

#ifdef USE_MPI
  MPI_Init(argc, &argv);
  MPI_Comm_size(MPI_COMM_WORLD, &size);
  MPI_Comm_rank(MPI_COMM_WORLD, &rank);
#else
  size = 1;
  rank = 1;
#endif

	Initialize(rank);
	tStart = time(NULL); 
        printf("Starting process %d\n",rank);

  Set_Size();
 	Init_File(rank);									// initialize output file for the run
  Read_WEISSQ(SEQUENCE_FILE);
	Read_Count();	

	for (Year = FIRST_YEAR_INDEX; Year < N_YEARS-1; Year++)   // annual prediction loop; start in 1993 (except AFG-6 starts in 1994)
	{
		fprintf(fres,"\n\n********************\n"); 
		fprintf(fres,"YEAR %4d\n",FIRST_YEAR+Year); 
		fprintf(fres,"********************\n\n"); 
		
		for (nex=0; nex<GA_EXPER; nex++)
		{
			printf("\n========\n\nYear %d  Experiment %d\n",FIRST_YEAR+Year,nex);

			Get_Template_Dates();
			if (fNoTemplates) {		// Get_Template_Dates failed
				printf("Insufficient templates found\n\n");
				fprintf(fres,"Insufficient templates found\n\n");
				fNoTemplates = FALSE;  // reset the flag
				continue;  // terminate this experiment
			}

			Estimate_GA();
			File_Results(nex);		
			Write_Forecast(TRUE);   // write the full forecast, then predictive forecast;
			Write_Forecast(FALSE);  //  see notes in the routine onthe use of a side-effect here;
		}													//  call using TRUE *must* precede call using FALSE
	} // Year

	tEnd = time(NULL);
	printf("\aFinished at %s\n",ctime(&tEnd));  
	printf("Elapsed time %8.2f minutes\n",(difftime(tEnd,tStart)/60.0));
	if (fres) 
	{
		fprintf(fres,"\nFinished at %s\n",ctime(&tEnd));  
		fprintf(fres,"Elapsed time %8.2f minutes\n",(difftime(tEnd,tStart)/60.0));
		fclose(fres);
	}

#ifdef USE_MPI
  MPI_Finalize();
#endif

} // main

//__________________________________________________________________________________
//																						main programs from Meyers and Whitson

void Train_HMM(int seq_length)
// create a hidden markov model from a series of symbol sequences using a randomly 
// generated model.
{
double 	min_delta_psum = MIN_DELTA; 

// char newfilename[100];	
    
  Init_Random_HMM();    											// initialize model randomly
  batch_train(seq_length,min_delta_psum);    // train model on data
  																						// dump the resulting model to a file
//  sprintf(newfilename,"%s.train.hmm",train_file);
//  dump_model(newfilename);
	  
} // Train_HMM


//__________________________________________________________________________________
//																																				utilities

void Initialize (int rank )
//
{
int seed;

	if (SEQ_SIZE>MAX_LENGTH) 
		{ printf("Error: SEQ_SIZE (%4d)>MAX_LENGTH (%4d)",SEQ_SIZE, MAX_LENGTH); exit(0); }

	Calc_Memory();
	
	srand(time(NULL)*(rank+1));												// randomly seed the RNG

	for (seed=0; seed < MAX_SAVE; seed++) 
	{
		aSaveAlpha[seed] = 1.0;   // flag to show cell does not contain real data
	}
	
} // Initialize

void Calc_Memory( void )
// Calculates the total memory requirements of the major arrays
// ### not sure if this includes the Viterbi arrays... <8.8>
{
long total;
	total = (MAX_STATES * sizeof(state)) * (MAX_SAVE + 3);// aStates
	total += sizeof(int) * MAX_STRINGS*(MAX_LENGTH+1);    // strings, string_len
	total += sizeof(arTrellState) * 6;										// alpha, beta, xi, gammaF
	total += sizeof(arTrell);															// scaling_factors
	total += sizeof(arState) * 5;													// a_numer, a_denom, b_denom
	total += sizeof(arStateSymb);													// b_numer
	total += MAX_WEIS * (sizeof(int) + sizeof(int));				// aWEIS arrays
	total += (MAX_STATES * sizeof(state)) * (2 * GA_VECT);	// GA vectors
	total +=  sizeof(float) * (3 * MAX_WEEKS);						// prob and aCount arrays
	total += (MAX_STATES * sizeof(state)) * (2 * MAX_SAVE + 4);	// XC vectors
	printf("Main array memory required: %ld bytes\n",total);
} // Calc_Memory

//__________________________________________________________________________________
//																																		HMM procedures



void Init_Saved_HMMGA(int top)
// set max_states, max_symbols and model parameters from file
{
int s;
	 assert((top>=0) && (top < 2*GA_VECT));
   for (s=0; s < max_states; s++) 
			aStates[s] = aGAFinal[s][top];
} // Init_Saved_HMMGA

void Set_Size(void)
// set globals max_symbols and max_states
{
  max_symbols = RUN_SYMBOLS;
  max_states = RUN_STATES;
} //Set_Size

void Init_Random_HMM(void)
// create hmm, initialize model parameters with random values.
{
  Set_Size(); 						// set size globals

  rnd_init_hmm_probs();  // rnd initialize the trans and emit probs

//  printf("\nUsing a random model...\n");
//  show_probs();
} //Init_Random_HMM


//__________________________________________________________________________________
//																																				Utilities

double UnifRnd( void )
// Uniform[0,1] random number generator
{
return  ((double)rand())/RAND_MAX;
}

long Julian (char *s)
// Converts dates of form YYMMDD to Julian serial beginning 0 = 1 Jan 1904
// (which matches the Excel convention) 
{
int 	mo,dy;
long	yr, ij;

  for (mo=0; mo<7; mo++) if (s[mo] == ' ') s[mo] = '0';			// convert blanks to zeros
  if (s[0] == '0') s[0] = ':';															// Y2K adjustment for the 00 to 09 dates
  
																		//Break date out of the string and convert to integers 
	yr = (short)s[0] * 10 + (short)s[1] - 532;			// 532 substracts the 1904 base from yr
	mo = (short)s[2] * 10 + (short)s[3] - 528;
	dy = (short)s[4] * 10 + (short)s[5] - 528;
						
	ij =  yr * 365;
	ij += ((yr+3) / 4) + Month[mo] + (dy-1);				// ((yr+3)/4) is the leap-year adjustment
	if (((yr % 4) == 0) && (mo > 2)) ij++;	// adjust for month within a leap-year
	return ij;
} // Julian


//__________________________________________________________________________________
//																										Meyers and Whitson estimation

void batch_train( int seq_length, double min_delta_psum) 
// train hmm from the existing sequence file until the sum of the probability
// changes is less than min_delta_psum.

// *** Changes by Emdad Khan: the code for the unequal length strings
// is changed to batch mode. It does not work in nonbatch mode.
// The equations given in Rabiner and elsewhere also indicates batch 
// mode training. Yi[] and yi[][] are used to add the batch mode update

// *** Changes in hmm.pa: sequences are now read a single time in a main();
//     seq_length is now passed instead of a file name
{
#define REVERSE_THRESHOLD 3
#define MAX_ITERATIONS    128

  int str;
  int		nitr=0;			// ### trying to track down an infinite loop here

  // local arrays and integers used in the case with strings of unequal lengths
  arState 			y1, y2, y3, y4, y6; 
  arStateSymb		y5;
  int i,k;
  int kreverse = 0;							// check number of alpha-reverse conditions, exit if >REVERSE_THRESHOLD

  double prob_sum=999999.0;						 // insure this will get reset
  double alpha_tot,last_alpha_tot;

  trellis_width = seq_length+1;  							// define trellis dimensions

																										  // loop until local min in probs
  for (str=0,alpha_tot=0.0; str<num_strings; str++)
    alpha_tot+=test(strings[str],string_len[str]);
#if fWRITE_ALPHA
  printf("\nlog(alpha)\tprob_sum\n");
  printf("----------\t--------\n");
  printf("%f\n",alpha_tot);
#endif
																					  // do batch update for each case:
  while (prob_sum > min_delta_psum) 
  {

    if (++nitr > MAX_ITERATIONS) 			// ### check if this is where the infinite loop is hit
    {
    	if (fres) fprintf(fres,"!! MAXIMUM ITERATION LIMIT EXCEEDED!\n");
    	printf("\a!! MAXIMUM ITERATION LIMIT EXCEEDED!\n");
    	break;
    }

    last_alpha_tot = alpha_tot;
    for (i=0 ; i < max_states; i++) 
    {
      y1[i] = 0;
      y2[i] = 0;
      y3[i] = 0;
      y4[i] = 0;
			y6[i] = 0;

			for (k=0; k < max_symbols; k++) 
			{
			  y5[i][k] = 0;
			}
    }

	  for (str=0; str<num_strings; str++) 
	  {																							// sum estimates across all training cases
			set_ab_counts(strings[str],string_len[str]);

			for (i=0; i < max_states; i++)
			{
			  y1[i] += a_numer_sum_recur[i];
		    y2[i] += a_numer_sum_next[i]; 
			  y3[i] += a_numer_sum_prev[i]; 
			  y4[i] += a_denom_sum[i]; 
			  y6[i] += b_denom_sum[i];

			  for (k=0; k < max_symbols; k++) 
			  {
			    y5[i][k] += b_numer_sum[i][k]; 
			  }
			}
   	}  // end of for str= loop

    // Transferring data to proper arrays/variables to be used 
    // in the reestimate routine
    for (i=0; i < max_states; i++) 
    {
			a_numer_sum_recur[i] = y1[i];
			a_numer_sum_next[i]  = y2[i]; 
			a_numer_sum_prev[i]	 = y3[i]; 
			a_denom_sum[i]			 = y4[i]; 
			b_denom_sum[i] 			 = y6[i];

			for (k=0; k < max_symbols; k++) 
			{
			  b_numer_sum[i][k] = y5[i][k]; 
			}
    }

    // restimate routine initializes the prob_sum
    prob_sum = reestimate();

    for (str=0, alpha_tot=0.0; str<num_strings; str++) 
				alpha_tot+=test(strings[str],string_len[str]);     

#if fWRITE_ALPHA
  printf("%f\t%f\n",alpha_tot,prob_sum);				// running output of alpha values
#endif
  if (alpha_tot < last_alpha_tot) 
  {
#if fWRITE_ALPHA
    printf("\tWARNING: alpha values reversed!\n"); // should not occur! (but it does!!...)
#endif
    kreverse++;
    if (kreverse > REVERSE_THRESHOLD) break;   // bail out if this occurs too often
   }
 } // while 

//  printf("\nDone training model %s  ...\n",filename);
  curAlpha = alpha_tot;
#if fWRITE_FILE
  printf("alpha_tot = %10.2f\n",alpha_tot);
#endif
  // show_probs();

} //batch_train 


void check_save(void)
// check whether we should save the model.
{
int ki, k, s;

  for (k=0; k < MAX_SAVE; k++) 
  {
    if ((aSaveAlpha[k] > 0.0) | (curAlpha > aSaveAlpha[k]))
    {	
     	printf("               Improved curAlpha(%1d) = %10.2f\n",k,curAlpha);

		 	if (k != MAX_SAVE-1)									// move remaining values down one slot
	    { for (ki=MAX_SAVE-1; ki > k; ki--) 
		    { aSaveAlpha[ki] =  aSaveAlpha[ki-1];
		      for (s=0; s < max_states; s++) 
						aSaveSt[s][ki] = aSaveSt[s][ki-1];
				}
			}
																						// put new information in slot k
		 aSaveAlpha[k] = curAlpha;
     for (s=0; s < max_states; s++) 
				aSaveSt[s][k] = aStates[s];
		break;																	// we're done		
		}
  }
} // check_save(void)


void show_probs(void)
// output the model parameters.
{
int k,s;

  printf("\nstates: %d\nsymbols: %d\n",max_states,max_symbols);
  for (s=0; s < max_states; s++) 
  {
    printf("%8.4f ",set_prev_trans(&aStates[s],-1.0));
    printf("%8.4f ",set_recur_trans(&aStates[s],-1.0));
    printf("%8.4f ",set_next_trans(&aStates[s],-1.0));
    for (k=0; k < max_symbols; k++)
      printf("%8.4f ",set_symb_out(&aStates[s],k,-1.0));
    printf("\n\n");
  }
  printf("\n");
} // show_probs(void)

void dump_model(char *fname)
// output the model parameters to a file.
{
int 		k,s;
FILE 		*fout;

	fout=fopen(fname,"w");
  if (!fout) 
  {
    printf("ERROR: Couldn't create file for model dump.");
    exit(-1);
  }

  fprintf(fout,"states: %d\n",max_states);
  fprintf(fout,"symbols:  %d\n",max_symbols);
  for ( s=0; s < max_states; s++) 
  {
    fprintf(fout,"%8.4f\t",set_prev_trans(&aStates[s],-1.0));
    fprintf(fout,"%8.4f\t",set_recur_trans(&aStates[s],-1.0));
    fprintf(fout,"%8.4f\t",set_next_trans(&aStates[s],-1.0));
    for ( k=0; k < max_symbols; k++)
      fprintf(fout,"%8.4f\t",set_symb_out(&aStates[s],k,-1.0));
    fprintf(fout,"\n\n");
  }
  fprintf(fout,"\n");
  fclose(fout);
  printf("\nDumped model to file ==> %s\n",fname);
}

#if FALSE
void dump_seq(FILE *seq, int length)
// generate a random sequence according to model parameters.
{
int 		j, cur_state,k,symbol;
double 	prob_trans,prob_out,accum;

  cur_state=0;
  for (j=0; j < length; j++) 
  {
    																										// fill in random
    prob_trans = UnifRnd();    
    prob_out = UnifRnd();    
    if (prob_trans < set_recur_trans(&aStates[cur_state],-1.0))
    {
      accum = 0.0;
      symbol=max_symbols-1;
      for (k=0; k < max_symbols; k++) 
      {
				if (prob_out < set_recur_out(&aStates[cur_state],k,-1.0) + accum) 
				{
				  symbol = k;
				  break;
				}
				else
				  accum += set_recur_out(&aStates[cur_state],k,-1.0);
			}
      fprintf(seq,"%2d ",symbol);
    }
    else 
    {
      accum = 0;
      symbol=max_symbols-1;
      for ( k=0; k < max_symbols; k++) 
      {
				if (prob_out < set_next_out(&aStates[cur_state],k,-1.0) + accum) 
				{
				  symbol = k;
				  break;
				}
				else
				  accum += set_next_out(&aStates[cur_state],k,-1.0);
			}
	    fprintf(seq,"%2d ",symbol);
	    cur_state++;
    }
  }

  if (cur_state < max_states-1)  printf("WARNING: Sequence did not reach final state!\n");
  fprintf(seq,"\n");

} // dump_seq
#endif

//__________________________________________________________________________________
//																															model initialization

#if fLR

void rnd_init_hmm_probs(void)
// randomly initialize the model parameters for LR model
{
double 	pr;
arSymb	prob_syms;
int 		i,j;
																				// set aij
  for (i=1; i < max_states-1; i++)
  {
    pr = UnifRnd();    
    set_recur_trans(&aStates[i],pr);
    set_next_trans(&aStates[i],1-pr);
	}
																				// set terminal state
  i = max_states-1;
  set_recur_trans(&aStates[i],1.0);
  set_next_trans(&aStates[i],0.0); 			// no next transition

  for (i=0; i < max_states; i++)			// set b-ik
  {
    rnd_sym_probs(prob_syms);
    for (j=0; j < max_symbols; j++)
    {
			set_symb_out(&aStates[i],j,prob_syms[j]);
		}
  }

} // rnd_init_hmm_probs for LR

#else

void rnd_init_hmm_probs(void)
// randomly initialize the model parameters
{
double 	pr, pp, pn, prob_sum;
arSymb	prob_syms;
int 		i,j;
																				// set aij
	if (fCIRC)
	{
	  for (i=0; i < max_states; i++)
	  {
	    pp = UnifRnd();    
	    pr = UnifRnd();    
	    pn = UnifRnd();
	    prob_sum = pp + pr + pn;    
	    set_prev_trans(&aStates[i],pp/prob_sum);
	    set_recur_trans(&aStates[i],pr/prob_sum);
	    set_next_trans(&aStates[i],pn/prob_sum);
		}
	}
	else 
	{
	  for (i=1; i < max_states-1; i++)
	  {
	    pp = UnifRnd();    
	    pr = UnifRnd();    
	    pn = UnifRnd();
	    prob_sum = pp + pr + pn;    
	    set_prev_trans(&aStates[i],pp/prob_sum);
	    set_recur_trans(&aStates[i],pr/prob_sum);
	    set_next_trans(&aStates[i],pn/prob_sum);
		}
																					// set state 0
	  pr = UnifRnd();    
	  set_prev_trans(&aStates[0],0.0); 			// no prev transition
	  set_recur_trans(&aStates[0],pr);
	  set_next_trans(&aStates[0],1.0-pr);
																					// set terminal state
	  i = max_states-1;
	  pr = UnifRnd();    
	  set_prev_trans(&aStates[i],1.0 - pr);
	  set_recur_trans(&aStates[i],pr);
	  set_next_trans(&aStates[i],0.0); 			// no next transition
	}																	

  for (i=0; i < max_states; i++)			// set b-ik
  {
    rnd_sym_probs(prob_syms);
    for (j=0; j < max_symbols; j++)
    {
			set_symb_out(&aStates[i],j,prob_syms[j]);
		}
  }

} // rnd_init_hmm_probs
#endif

void rnd_sym_probs(double prob_syms[])
// set output symbol probabilities.
{
double 	prob_sum = 0.0;
int 		s;

  for (s=0; s < max_symbols; s++) 
  {
    prob_syms[s] = UnifRnd();
    prob_sum += prob_syms[s];
  }

  for (s=0; s < max_symbols; s++)  // normalize so they add to one
    prob_syms[s] = prob_syms[s] / prob_sum;

} // rnd_sym_probs


//__________________________________________________________________________________
//																																			calculations

void rescale_alphas(int col)
// rescale alpha values (from Rabiner).
{ 
int i;

  scaling_factors[col] = 0.0;
  for ( i=0; i < max_states; i++) 
  {
    scaling_factors[col] += alpha[col][i];
  }

  for (i=0; i <  max_states; i++)        // rescale all the alpha's and smooth them
  {
    alpha[col][i] /= scaling_factors[col];
  }
} // rescale_alphas

void rescale_betas(int col)
// rescale beta values after rescaling alphas.
{ 
int i;

  for ( i=0; i < max_states; i++)  			// rescale all the beta's w alpha's factors
  {
    beta[col][i] /= scaling_factors[col];
  }
} // rescale_betas

double alpha_F(int* symbol_array, int seq_length)
// calculate alpha trellis and return (and set) final sequence probability
{ 
int 		t,j;
int 		m = max_states-1;
double 	acirc = 0.0;					// extra term for circular model

  seq_length=(seq_length<0?trellis_width-1:seq_length);

  // clear the trellis and set the first column
  for (t=0; t < trellis_width; t++)
    for (j=0; j < max_states; j++) 
    {
      alpha[t][j] = 0.0;
    }
  alpha[0][0] = 1.0;  // first state starts with alpha = 1.0
  rescale_alphas(0);  // first col rescale is div by one

  // calculate the alphas for rest of cols
  for (t = 0; t < seq_length; t++) 
  { 		
    if (fCIRC) 
    	acirc = alpha[t][m] * set_next_trans(&aStates[m],-1.0);
    	
    alpha[t+1][0] = (
								      (alpha[t][0] * set_recur_trans(&aStates[0],-1.0) ) +
											(alpha[t][1] * set_prev_trans(&aStates[1],-1.0) )  +
											acirc
									  ) *
					  				set_symb_out(&aStates[0],symbol_array[t],-1.0);

    for (j=1; j < m; j++) 
    {
      alpha[t+1][j] = (
									      (alpha[t][j] * set_recur_trans(&aStates[j],-1.0) ) +
												(alpha[t][j-1] * set_next_trans(&aStates[j-1],-1.0) ) +
												(alpha[t][j+1] * set_prev_trans(&aStates[j+1],-1.0) )
										  ) *
						  				set_symb_out(&aStates[j],symbol_array[t],-1.0);
    }

    if (fCIRC) 
    	acirc = alpha[t][0] * set_prev_trans(&aStates[0],-1.0);

    alpha[t+1][m] = (
								      (alpha[t][m] * set_recur_trans(&aStates[m],-1.0) ) +
											(alpha[t][m-1] * set_next_trans(&aStates[m-1],-1.0) ) +
											acirc
									  ) *
					  				set_symb_out(&aStates[m],symbol_array[t],-1.0);
   rescale_alphas(t+1);
  }
										// compute probability of complete sequence
										// ### is this accurate given the scaling??
  prob_seq = 0.0;
  for (j=1; j < m; j++)
  	prob_seq += alpha[seq_length][j];
  	
  return (prob_seq);
} // alpha_F

double beta_I(int* symbol_array, int seq_length)
// calculate beta trellis and return initial beta value
{
int 		t,j;
int 		m = max_states-1;
double 	acirc = 0.0;					// extra term for circular model
	
 seq_length=(seq_length<0?trellis_width-1:seq_length);     

																							 // clear the trellis and set the last column
 for ( t=0; t < trellis_width; t++)
   for ( j=0; j < max_states; j++) 
   {
		 beta[t][j] = 0.0;
   }

	for ( j=0; j < max_states; j++) 
	{
		beta[seq_length][j]	 = 1.0;  		// final betas are 1.0 
	}
// rescale_betas(seq_length); 								// this is really div by 1 so no effect

 																							// begin setting all the cols except last one
 for ( t = seq_length-1; t >= 0; t--) 
 {
    if (fCIRC) 
    	acirc =  ( beta[t+1][m] * set_prev_trans(&aStates[m],-1.0) *
				         set_symb_out(&aStates[m],symbol_array[t],-1.0)	);

	  beta[t][0] = ( 
		 								beta[t+1][0] * set_recur_trans(&aStates[0],-1.0) *
				            set_symb_out(&aStates[0],symbol_array[t],-1.0)
				          ) + (
				     				beta[t+1][1] * set_next_trans(&aStates[0],-1.0) * 
						     	  set_symb_out(&aStates[1],symbol_array[t],-1.0) 
				          ) + acirc; 
				          
   for ( j=1; j < m; j++) 
   {
		 beta[t][j] = ( 
		 								beta[t+1][j] * set_recur_trans(&aStates[j],-1.0) *
				            set_symb_out(&aStates[j],symbol_array[t],-1.0)
				          ) + (
				     				beta[t+1][j+1] * set_next_trans(&aStates[j],-1.0) * 
						     	  set_symb_out(&aStates[j+1],symbol_array[t],-1.0) 
 				          ) + (
				     				beta[t+1][j-1] * set_prev_trans(&aStates[j],-1.0) * 
						     	  set_symb_out(&aStates[j-1],symbol_array[t],-1.0)
						     	); 
  }

  if (fCIRC) 
  	acirc =  ( beta[t+1][0] * set_next_trans(&aStates[m],-1.0) *
			         set_symb_out(&aStates[0],symbol_array[t],-1.0)	);

	 beta[t][m] = ( 
	 								beta[t+1][m] * set_recur_trans(&aStates[m],-1.0) *
			            set_symb_out(&aStates[m],symbol_array[t],-1.0)
			          ) + (
			     				beta[t+1][m-1] * set_prev_trans(&aStates[m],-1.0) * 
					     	  set_symb_out(&aStates[m-1],symbol_array[t],-1.0)
				        ) + acirc; 

   rescale_betas(t);
 }

 return (beta[0][0]);
} // beta_I


void compute_gamxi(int* symbol_array, int seq_length)
// compute xi and gamma values and fill tables for entire trellis.
//
// ### NOTE: It should be possible to pull the accum out of the loop,
// as this should be equal for each t? (or does rescaling screw that up?)
// Should check on that.  The divisor is Pr(O|lam)
// Also: Given that I've initialized everything to zero, I think some of 
// those parameter sets to zero are redundant.
{
int i,t,j;
int 	m = max_states-1;
double prob_ck=0;		// ###
double accum;
 
  seq_length=(seq_length<0?trellis_width-1:seq_length);

 																		 	// clear xis
  for ( t=0; t < seq_length; t++ )
    for ( j=0; j < max_states; j++ ) 
    {
			xi_prev[t][j] = 0.0;
			xi_recur[t][j] = 0.0;
			xi_next[t][j] = 0.0;
			gammaF[t][j] = 0.0;
    }

 																			 // calc the xi and gamma tables
  for ( t = 0; t < seq_length; t++ ) 
  {
    accum = 0;
    for ( i=0; i < max_states; i++ ) 
    {
      if (i>0)
      { xi_prev[t][i] = ( alpha[t][i] *
			  									set_prev_trans(&aStates[i],-1.0) *
			  									set_symb_out(&aStates[i-1],symbol_array[t],-1.0) *
			  									beta[t+1][i-1] );
			 }
			 else if (fCIRC)
	      { xi_prev[t][0] = ( alpha[t][0] *
				  									set_prev_trans(&aStates[i],-1.0) *
				  									set_symb_out(&aStates[m],symbol_array[t],-1.0) *
				  									beta[t+1][m] );
			 }
			 else xi_prev[t][0] = 0.0;

      if (i<m)
      {
	      xi_next[t][i] = ( alpha[t][i] *
			  									set_next_trans(&aStates[i],-1.0) *
			  									set_symb_out(&aStates[i+1],symbol_array[t],-1.0) *
			  									beta[t+1][i+1] );
			}
			else if (fCIRC)
      { xi_next[t][m] = ( alpha[t][m] *
			  									set_next_trans(&aStates[m],-1.0) *
			  									set_symb_out(&aStates[0],symbol_array[t],-1.0) *
			  									beta[t+1][0] );
			}
			else xi_next[t][m] = 0.0;
			
      xi_recur[t][i] = (	alpha[t][i] * 
		   										set_recur_trans(&aStates[i],-1.0) * 
		  									 	set_symb_out(&aStates[i],symbol_array[t],-1.0) *
		   										beta[t+1][i] );
				   									
		  gammaF[t][i] = xi_prev[t][i] + xi_recur[t][i] + xi_next[t][i];
		  accum +=  gammaF[t][i];
    } // for i

    for ( i=0; i < max_states; i++ )    // now adjust for the sum
		{
		 xi_prev[t][i] /= accum;
		 xi_recur[t][i] /= accum;
		 xi_next[t][i] /= accum;
		 gammaF[t][i] /= accum;
		}
  }  // for t
    
} // compute_gamxi


double a_numer(int i, int j, int seq_length)
// this is the numerator of the a_ij reestimate and denom of b_ij reestimate,
// this func assumes gammas have been calc'd.
// ### NOTE: reverse the loop/if once this is working
{
double sum = 0.0;
int t;

  seq_length=(seq_length<0?trellis_width-1:seq_length);

  if ((j<0) || (j>=max_states))
  { if (fCIRC)
  	{	if ( j < 0 )
	 		{
			  assert(i == 0);		// ###
			  for( t=0; t < seq_length-1; t++)
			  	 sum += xi_prev[t][i];
	 		}
	 		else 
	 		{
			  assert(i == max_states-1); // ###
			  for( t=0; t < seq_length-1; t++)
			  	 sum += xi_next[t][i];
	 		}
  	}
  	return sum;
	} 


  for( t=0; t < seq_length-1; t++)   // -1 is consistent with Rabiner
    {
      if (i==j) 
      {
				sum += xi_recur[t][i];
      }
      else if ( j == i+1) 
      {
				sum += xi_next[t][i];
      }
      else if ( j == i-1) 
      {
				sum += xi_prev[t][i];
      } else 
      {
				printf("WARNING: xi_next[t][%d] shouldn't be requested\n",i);
				break;
      }
    }
  return sum;
} // a_numer


double a_denom(int i, int seq_length)
// this is the denominator of the a_ij reestimate.
{ 
int t;
double sum = 0.0;

  seq_length=(seq_length<0?trellis_width-1:seq_length);
	
  for( t=0; t < seq_length-1; t++)
      sum += gammaF[t][i];

  return sum;
} // a_denom

// 
double b_numer(int i, int sym, int *symbol_array, int seq_length)
// this is the numerator of the b_i,sym reestimate.
{	
int t;
double sum = 0.0;

  seq_length=(seq_length<0?trellis_width-1:seq_length);
	
  for( t=0; t < seq_length; t++) 
  {
   if ( symbol_array[t] == sym ) 
    {
			 sum += gammaF[t][i];
    }
  }
return sum;
} // b_numer

double b_denom(int i, int seq_length)
// this is the denominator of the b_i,sym reestimate.
{ 
int t;
double sum = 0.0;

  seq_length=(seq_length<0?trellis_width-1:seq_length);
	
  for( t=0; t < seq_length; t++)
      sum += gammaF[t][i];

  return sum;
} // b_denom


double set_ab_counts(int symbol_array[], int seq_length)
// set ab count from one string.
{
double 	alpha_tot;
int 		i,k;
double prob_ck=0;

#if FALSE
  for (i=0; i<seq_length; i++)		//###
  {
  	if ((symbol_array[i]<0) || (symbol_array[i]>max_symbols ))
  	    {  printf("Symbol error\n"); }

  }
// end ###
#endif

  alpha_tot = alpha_F(symbol_array,seq_length);		// fill the alpha matrix
  
  beta_I(symbol_array,seq_length);  							// fill the beta matrix

  compute_gamxi(symbol_array,seq_length);  				// fill the xi and gamma matrices

  for (i=0; i < max_states; i++) 
  {
	  a_numer_sum_prev[i] = a_numer(i,i-1,seq_length);
	  a_numer_sum_recur[i] = a_numer(i,i,seq_length);
	  a_numer_sum_next[i] = a_numer(i,i+1,seq_length);

	  a_denom_sum[i] = a_denom(i,seq_length);
	  b_denom_sum[i] = b_denom(i,seq_length);

//	  prob_ck=0;//###
	  for ( k=0; k < max_symbols; k++) 
	  {
	    b_numer_sum[i][k] = b_numer(i,k,symbol_array,seq_length);
//###	    prob_ck += b_numer_sum[i][k]/b_denom_sum[i];
	  }
//    if (fabs(1.0 - prob_ck) >0.01) 		// ###
//    {  printf("Set_ab_counts prob_ck error\n"); } // ###
  }

  return(alpha_tot);
} // set_ab_counts

double reestimate()
// reestimate parameters after calculating ab_counts
{
  double prob, diff_sum = 0.0;
	int i,k;
//	double check_prob;
	
  // loop all the transition probs

  for ( i=0; i < max_states; i++)
  {
																										    // do all the prob transitions except last one
    prob = a_numer_sum_recur[i] / a_denom_sum[i];
    diff_sum += fabs(prob - set_recur_trans(&aStates[i],-1.0));
    set_recur_trans(&aStates[i],prob);
 																											 // do all the next transitions
    prob = a_numer_sum_next[i] / a_denom_sum[i];
    diff_sum += fabs(prob - set_next_trans(&aStates[i],-1.0));
    set_next_trans(&aStates[i],prob);

 																											// do all the prev transitions
    prob = a_numer_sum_prev[i] / a_denom_sum[i];
    diff_sum += fabs(prob - set_prev_trans(&aStates[i],-1.0));
    set_prev_trans(&aStates[i],prob);
																												// compute observation probs
//    check_prob=0.0; //###
    for ( k=0; k < max_symbols; k++) 
    {
			prob = b_numer_sum[i][k] / b_denom_sum[i];
			diff_sum += fabs(prob -	set_symb_out(&aStates[i],k,-1.0));
			set_symb_out(&aStates[i],k,prob);
//###			check_prob += prob;
    }
//###    if (fabs(1.0 - check_prob) >0.01) 
//###    {  printf("Checkprob error\n"); }

  }

  return diff_sum;
} // reestimate()

//
double test(int* string, int seq_length)
// test model for a single string.
{
  double log_alpha;
	int t;
	
  seq_length=(seq_length<0?trellis_width-1:seq_length);

  // fill alpha trellis and scaling coeffs
  log_alpha=log(alpha_F(string,seq_length));

  // calc log probability from coeffs
  if (scaling_factors[0]>0.0) 
  { // if scaling_factors set
    log_alpha=0.0;
    for ( t=0; t < seq_length+1; t++)
      log_alpha += log(scaling_factors[t]);
  }

//	Viterbi(0);						// use only if computing qstar

  return log_alpha;
} // test

//__________________________________________________________________________________
//																																	aState functions

double set_symb_out(state *pState, int symbol, double prob)
// set/return the probability of generating a particular symbol for pState, 
// (smooth with MIN_PROB)
{
  if (prob != -1.0) 
  {
    pState->symb_out[symbol] = prob;
  }
  if (pState->symb_out[symbol] < MIN_PROB)
    pState->symb_out[symbol] = MIN_PROB;
  return pState->symb_out[symbol];
} // set_symb_out

double set_recur_trans(state *pState,double prob)
// set/return the probability of making a recurrent transition.
// (smooth with MIN_PROB)
{
  if (prob != -1.0) 
  {
    pState->recur_trans = prob;
  }
  if (pState->recur_trans < MIN_PROB)
    pState->recur_trans = MIN_PROB;
  return pState->recur_trans;
} // set_recur_trans

double set_next_trans(state *pState,double prob)
// set/return the probability of making transitioning to the next state.
// (smooth with MIN_PROB)
{
  if (prob != -1.0) 
  {
    pState->next_trans = prob;
  }
  if (pState->next_trans < MIN_PROB)
    pState->next_trans = MIN_PROB;
  return pState->next_trans;
} // set_next_trans

double set_prev_trans(state *pState,double prob)
// set/return the probability of making transitioning to the next state.
// (smooth with MIN_PROB)
{
 #if fLR
    pState->prev_trans = 0;
#else
 if (prob != -1.0) 
  {
    pState->prev_trans = prob;
  }
  if (pState->prev_trans < MIN_PROB)
    pState->prev_trans = MIN_PROB;
#endif
  return pState->prev_trans;
} // set_prev_trans

//__________________________________________________________________________________
//																																Viterbi algorithm
// 																							(compute most likely state sequence)

void Viterbi( int index)
// compute MLE state sequence for string[index] using	Viterbi algorithm from Rabiner
// Note: this does not do any rescaling of the deltas, so on a long sequence these
//       could underflow.

{
int 		j, kt;
double 	recur_del, next_del, prev_del,dmax;			// values of delta for probs
int			maxi;											// state that produced max psi 

// Initialization
	for (j=0; j < max_states; j++)
	{
		delta[0][j] = set_symb_out(&aStates[j],strings[index][0],-1.0) ;
		psi[0][j] = 0;
	}
	
// fill trellis using recursive formula 
	for (kt=1; kt < string_len[index]; kt++)
	{
		for (j=0; j < max_states; j++)			// compute delta and psi for each state
		{
			recur_del = delta[kt-1][j]*set_recur_trans(&aStates[j],-1.0) ;
			if (j>0)
				next_del = delta[kt-1][j-1]*set_next_trans(&aStates[j-1],-1.0) ;
			else 
				next_del = 0;
			if (j<max_states - 1)
				prev_del = delta[kt-1][j+1]*set_prev_trans(&aStates[j+1],-1.0) ;
			else 
				prev_del = 0;
																				// find maximum
			if (recur_del > next_del)
			{
				dmax = recur_del;
				maxi = j;
			}
			else 
			{
				dmax = next_del;
				maxi = j-1;
			}
			if (prev_del > dmax)
			{
				dmax = prev_del;
				maxi = j+1;
			}
		
			delta[kt][j] = dmax * set_symb_out(&aStates[j],strings[index][kt],-1.0) ;
			psi[kt][j] = maxi;
		}		
	}			// kt

// backtrack through psi to get the optimal sequence

	kt = string_len[index]-1;
	maxi=0;
	recur_del = delta[kt][0];							// use recur_del to store maximum
	
	for (j=1; j < max_states; j++)			// find maximum delta[T]
		if (recur_del < delta[kt][j])
		{
			recur_del = delta[kt][j];
			maxi = j;
		}
	qstar[kt] = maxi;
	
	for (kt=string_len[index]-1; kt > 0 ; kt--)
	{
		qstar[kt-1] = psi[kt][qstar[kt]];		
	}
} // Viterbi

//__________________________________________________________________________________
//																									Proportion of state calculation

double qprop(int state, int length)
// return proportion of qstar that have the value state
// note that this is currently summing over a set of two states
{
int ka, kount=0;

	for (ka= 0; ka<length; ka++)
//		if (qstar[ka] == state) kount++;
		if ((qstar[ka] == state) || (qstar[ka] == state+1)) kount++;
	return ((double)kount/length);
} // qprop


//__________________________________________________________________________________
//																												Continuous WEIS comparison

void Read_WEISSQ(char *sfile)
// read a WEISSQ file into aWEIS arrays.  Only saves the days that are
// >= BASE_DATE
{
FILE 	*fin;
int		ka = 0;

	printf("\nReading sequence %s\n",sfile);
	fin = fopen(sfile,"r");
	if (!fin) { printf("Error: can't open %s",sfile); exit(0); }

	while ((!feof(fin)) && (ka<MAX_WEIS))
	{	fscanf(fin,"%ld %d",&aWEIS_Date[ka],&aWEIS_Evt[ka]);
	  if (aWEIS_Date[ka] >= BASE_DATE) ka++; }
	
	if (ka >= MAX_WEIS) { printf("Error: file length exceeds MAX_WEIS in %s",sfile); exit(0); }

	NumWEIS=ka;	
	fclose(fin);
} // Read_WEISSQ


void Estimate_HMM(void)
// MC estimation from existing set of strings
{
int nex;

	for (nex=0; nex < MAX_SAVE; nex++) aSaveAlpha[nex] = 1.0;   // flag to show cell does not contain real data

  Set_Size();
	for (nex = 0; nex < N_EXPER; nex++)
	{ 
		Train_HMM(SEQ_SIZE);
	  check_save();

		#if fWRITE_FILE
		printf("\nExper %3d\n",nex);
    for (ka=0; ka < MAX_SAVE; ka++) printf("%2d   %10.2f\n",ka,aSaveAlpha[ka]);  // debugging
	  printf("\n");
		#endif

	}
	#if fWRITE_FILE
	printf("\nModel estimation finished\n");
	#endif
} // Estimate_HMM

void Next_Month (char *sMonth)
// increment the month string
{
  sMonth[3]++;
  if (sMonth[3] > '9')
  { sMonth[3] = '0';
  	sMonth[2] = '1';
  }
  else if ((sMonth[3] == '3') && (sMonth[2] == '1'))
  { sMonth[3] = '1';
  	sMonth[2] = '0';
	  sMonth[1]++;
	  if (sMonth[1] > '9')
	  { sMonth[1] = '0';
	  	sMonth[0]++;
	  }
	}  
} // Next_Month

void Compute_Running_Prob (char *SourceCode, char *TargetCode, char ctype )
// computes alpha for 100 events prior to the end of each month
{
char 		sStart[] = "790801";
char 		sEnd[]	 = "990101";						// final month plus 1
//char 		sEnd[]	 = "970601";						// final month plus 1
int 		ka, kb, ijul; 
int			lastDate=SEQ_SIZE-1;
double	val, val2 = 0.0;
int			kpr = 0;

#if fWRITE_FILE
FILE		*fout; 												// output file to write alpha values
char  	fname[32];

	printf("\nComputing alphas for %3.3s > %3.3s\n",SourceCode,TargetCode);

	sprintf(fname,"WEIS.%3.3s>%3.3s.%c.alphas",SourceCode,TargetCode,ctype);
	fout = fopen(fname,"w");
	if (!fout) { printf("Error: can't open %s",fname); exit(0); }
#endif

// set globals
  trellis_width = SEQ_SIZE+1;  // define trellis dimensions
  max_symbols = MAX_SYMBOLS;
  string_len[0] = SEQ_SIZE;
	
	while (strcmp(sStart,sEnd))
	{
		ijul = Julian(sStart);
		while (( aWEIS_Date[++lastDate] < ijul) && (lastDate < NumWEIS)) {};  // find the date in the array
		
		kb=0;
		assert(lastDate - SEQ_SIZE>=0);
		assert(lastDate < MAX_WEIS);
		for (ka = lastDate - SEQ_SIZE; ka < lastDate; ka++)			// load previous MAX_SEQ events into strings
			strings[0][kb++] = aWEIS_Evt[ka];											// Note that this is overwriting the first template

    val = test(strings[0],SEQ_SIZE);												// compute alpha
    val2 = qprop(1,SEQ_SIZE);															// compute % of state i
//    val2 = ijul - aWEIS_Date[lastDate - SEQ_SIZE];					// compute # days in sequence

#if fWRITE_FILE
    printf("alpha for %s =  %10.4f  qprop =  %10.4f\n",sStart,val,val2);
    fprintf(fout, "%s\t %10.4f\t %10.4f\n",sStart, val,val2); // write values to file
#endif
    
		Next_Month(sStart);
	}	

	// restore the zero string
	for (ka=0; ka<SEQ_SIZE; ka++) strings[0][ka] = string_zero[ka];

#if fWRITE_FILE
  fclose(fout);
#endif

} // Compute_Running_Prob


//__________________________________________________________________________________
//																																genetic algorithm

void Evolve(int eon)
// does genetic selection, recombination and mutation on existing vectors
{

double	aSortVal[GA_VECT];						// sorted values
int 		aSortIdx[GA_VECT];						// sorted index
double	val;
int			ka, kb, idx;
int			pa, ma, kid, kidx;
int			last, xover;

																			// copy fit info into aSort*
	for (ka = 0; ka<GA_VECT; ka++)
	{
		aSortVal[ka] = aGAFit[ka];
		aSortIdx[ka] = ka;
	}
																		// select sort vectors by fit
	for (ka = 0; ka<GA_VECT-1; ka++)
	{
		val = aSortVal[ka];
		idx = ka;
		for (kb = ka+1; kb<GA_VECT; kb++)
			if (val < aSortVal[kb])
			{
				val = aSortVal[kb];
				idx = kb;
			}
		if (idx != ka) 		// then swap
		{
			aSortVal[idx] = aSortVal[ka];
			aSortVal[ka] = val;

			kb = aSortIdx[idx];
			aSortIdx[idx] = aSortIdx[ka];
			aSortIdx[ka] = kb;
		}
	}		// end ka sort loop
																		// record current top vector
	aHistory[eon].val = aSortVal[0];
	aHistory[eon].age = aGAAge[aSortIdx[0]];
	aHistory[eon].top = aSortIdx[0];
	
																				// put cumulative fitness into SortVal; this is used to select parents;
	last = (int)(GA_VECT * GA_SURVIVE);		// index of last survivor
	val = 0.0;
	for (ka = 0; ka<last; ka++)
	{
		val += aSortVal[ka];
		aGAAge[aSortIdx[ka]]++;					// update the age while we're at it...
	}
	
	assert(val!=0.0);
	aSortVal[0] /=val;
	for (ka = 1; ka<last-1; ka++) aSortVal[ka] = aSortVal[ka-1] + aSortVal[ka]/val;
	aSortVal[last-1] = 1.0;
	for (ka = 0; ka<last-1; ka++) assert(aSortVal[ka]>=0 && aSortVal[ka]<=1.0);  // ### debug
	
	
	
	for (kid = last; kid<GA_VECT; kid++)
	{
		kidx = aSortIdx[kid];								// get index of the vector being replaced
																				//select parents
		val = UnifRnd();
		pa = 0;
		while (aSortVal[pa]<val) pa++;
		
		do
		{
			val = UnifRnd();
			ma = 0;
			while (aSortVal[ma]<val) ma++;
			if (pa == ma) ma = -1;				// duplicate, so find another
		} while (ma<0);
																		// replace these with their refs in the original array
		pa = aSortIdx[pa];
		ma = aSortIdx[ma];
																		
		for (ka=0; ka<max_states; ka++)
		{
			if (UnifRnd()<GA_MUTATE) 
			{
				Mutate_TransProbs(kidx,ka);
				Mutate_TransProbs(kidx + GA_VECT,ka);
			}			
			else
			{ 														 // transfer probabilities from a parent
				idx = (UnifRnd()<0.5) ? pa : ma;

				aGAStart[ka][kidx].prev_trans = aGAStart[ka][idx].prev_trans;
				aGAStart[ka][kidx].recur_trans = aGAStart[ka][idx].recur_trans;
				aGAStart[ka][kidx].next_trans = aGAStart[ka][idx].next_trans;

				aGAStart[ka][kidx + GA_VECT].prev_trans = aGAStart[ka][idx + GA_VECT].prev_trans;
				aGAStart[ka][kidx + GA_VECT].recur_trans = aGAStart[ka][idx + GA_VECT].recur_trans;
				aGAStart[ka][kidx + GA_VECT].next_trans = aGAStart[ka][idx + GA_VECT].next_trans;
			}

			xover = UnifRnd() * max_symbols;	// select cross-over point
			idx = (UnifRnd()<0.5) ? pa : ma;  // select source of first part of vector
			for (kb=0; kb<xover; kb++)
			{ 			
				aGAStart[ka][kidx].symb_out[kb] = aGAStart[ka][idx].symb_out[kb];
				aGAStart[ka][kidx + GA_VECT].symb_out[kb] = aGAStart[ka][idx + GA_VECT].symb_out[kb];
			}

			idx = (idx == pa) ? ma : pa;  		// idx for second part of vector
			for (kb=xover; kb<max_symbols; kb++) 
			{			
				aGAStart[ka][kidx].symb_out[kb] = aGAStart[ka][idx].symb_out[kb];
				aGAStart[ka][kidx + GA_VECT].symb_out[kb] = aGAStart[ka][idx + GA_VECT].symb_out[kb];
			}

																					// mutate
																					// still need to mutate the probs
			for (kb=0; kb<max_symbols; kb++) 			
				if (UnifRnd()<GA_MUTATE) aGAStart[ka][kidx].symb_out[kb] = UnifRnd();

			val = 0.0;													// renormalize the high probabilities
			for (kb=0; kb<max_symbols; kb++) 			
				val +=aGAStart[ka][kidx].symb_out[kb];
			for (kb=0; kb<max_symbols; kb++) 			
				aGAStart[ka][kidx].symb_out[kb] /=val;
								
			for (kb=0; kb<max_symbols; kb++) 			
				if (UnifRnd()<GA_MUTATE) aGAStart[ka][kidx + GA_VECT].symb_out[kb] = UnifRnd();

			val = 0.0;													// renormalize the low probabilities
			for (kb=0; kb<max_symbols; kb++) 			
				val +=aGAStart[ka][kidx + GA_VECT].symb_out[kb];
			for (kb=0; kb<max_symbols; kb++) 			
				aGAStart[ka][kidx + GA_VECT].symb_out[kb] /=val;

		} // end ka loop

		aGAAge[kidx] = 0;			// this will trigger a calculation of fitness 

	}  // end kid loop
	
} // Evolve

void Mutate_TransProbs(int kidx, int ka)
//	generate new transition probabilites for an LRL model
{
double 	pr, pp, pn, prob_sum;

	assert((kidx>=0) && (kidx<2*GA_VECT));
	if (!ka)
	{	// set state 0
	  pr = UnifRnd();   
	  aGAStart[ka][kidx].prev_trans  = 0.0; 			// no prev transition
	  aGAStart[ka][kidx].recur_trans = pr;
	 	aGAStart[ka][kidx].next_trans  = 1.0-pr;
	}
	else if (ka == max_states-1)
	{	// set terminal state
	  pr = UnifRnd();    
	  aGAStart[ka][kidx].prev_trans  = 1.0-pr;
	  aGAStart[ka][kidx].recur_trans = pr;
	 	aGAStart[ka][kidx].next_trans  = 0.0; 			// no next transition
	}
	else			// all other states
	{	
		pp = UnifRnd();    
	  pr = UnifRnd();    
	  pn = UnifRnd();
	  prob_sum = pp + pr + pn;    
	  aGAStart[ka][kidx].prev_trans  = pp/prob_sum;
	  aGAStart[ka][kidx].recur_trans = pr/prob_sum;
	  aGAStart[ka][kidx].next_trans  = pn/prob_sum;
	}
} // Mutate_Probs

void Initialize_Vectors(void)
// set up initial vectors
{
int ka, kb;

	for (ka=0; ka<GA_VECT; ka++)
	{
		rnd_init_hmm_probs();							// randomize a new model
																			// put information into ka vector
		aGAFit[ka] = 0.0;   							// GA fitness
		aGAAge[ka] = 0;   								// GA age

		rnd_init_hmm_probs();							// randomize high model
		for (kb=0; kb < max_states; kb++) 
			aGAStart[kb][ka] = aStates[kb];

		rnd_init_hmm_probs();							// randomize low model
		for (kb=0; kb < max_states; kb++) 
			aGAStart[kb][ka+GA_VECT] = aStates[kb];
	}
} // Initialize_Vectors

void Estimate_GA(void)
// Main GA estimation loop
{
int eon,ka;
	Initialize_Vectors();
	for (eon = 0; eon < GA_MAX_EON; eon++)
	{ 
		for (ka=0; ka<GA_VECT; ka++)
			if (!aGAAge[ka])
			{		

				Estimate_Vector(ka);
			}

			Evolve(eon);
			if ((!eon) || (aHistory[eon-1].top != aHistory[eon].top))
				Write_GA_History(eon);
			else 		
				printf("%3d  %3d   %3d  %10.2f\n",eon,aHistory[eon].top, aHistory[eon].age,aHistory[eon].val - FITNESS_BASE); 
	}
		printf("Summary of GA Run:");
		Write_GA_History(eon-1);
} // Estimate_GA


void init_GA_probs(int idx)
// initialize the model parameters from aGAStart[idx]
{
int 		i,j;

  for (i=0; i < max_states; i++)
  {
    set_prev_trans(&aStates[i],aGAStart[i][idx].prev_trans);
    set_recur_trans(&aStates[i],aGAStart[i][idx].recur_trans);
    set_next_trans(&aStates[i],aGAStart[i][idx].next_trans);
    
	  for (j=0; j < max_symbols; j++)
	  {
			set_symb_out(&aStates[i],j,aGAStart[i][idx].symb_out[j]);
		}
	}
} // init_GA_probs

void Estimate_Vector(int idx)
// Fitness computed using paired high/low vector pairs
{
int kb;

 	Init_Template_HMM(aHigh);
 	init_GA_probs(idx);    									// initialize model from vector idx high
 	batch_train(SEQ_SIZE,MIN_DELTA);    		// train model on data
	for (kb=0; kb < max_states; kb++) 
		aGAFinal[kb][idx] = aStates[kb];
	Compute_Weekly_Prob(aHighProb);
		
 	Init_Template_HMM(aLow);
 	init_GA_probs(idx+GA_VECT);    					// initialize model from vector idx low
 	batch_train(SEQ_SIZE,MIN_DELTA);    		// train model on data
	for (kb=0; kb < max_states; kb++) 
		aGAFinal[kb][idx+GA_VECT] = aStates[kb];
	Compute_Weekly_Prob(aLowProb);

	aGAFit[idx] = Compare_High_Low();

} // Estimate_Vector (high/low pairs)

void Write_GA_History(int eon)
// write the aHistory array to screen
{
int ka;
	putchar('\n');
	for (ka=0; ka<=eon; ka++)
	if ((!ka) || (aHistory[ka-1].top != aHistory[ka].top))
		printf("%3d  %3d   %3d  %10.2f\n",ka,aHistory[ka].top, aHistory[ka].age, aHistory[ka].val - FITNESS_BASE); 
	putchar('\n');
} // Write_GA_History


void Init_File(int rank)
// initializes fres
{
time_t tStart = time(NULL);
int		 ka = tStart % 10000;
char	 name_fres[] = RESULTS_FILE;

	
// assign lag and type to name_fres

	if 	(FORECAST == 28) name_fres[10] = '1';
	else if  (FORECAST == 91) name_fres[10] = '3';
	else 		 name_fres[10] = '6';

#ifdef P_MODEL
	name_fres[9] = 'P';
#else	
#ifdef N_MODEL
	name_fres[9] = 'N';
#endif
#endif

	if (rank<10) name_fres[13] = 48 + rank;     // append rank
        else name_fres[13] = 55 + rank;

	sprintf(&name_fres[15], "%04d",ka);     // append time-based suffix to name_fres

	fres=fopen(name_fres,"w");
  if (!fres) 
  {
    printf("ERROR: Couldn't create file \"%s\"for model results",name_fres);
    exit(-1);
  }
	
  printf("Opened file %s\n",name_fres);
	fprintf(fres,"HIDDEN MARKOV MODEL, MONTE CARLO OPTIMIZATION\n\n");
	fprintf(fres,"Run initiated: %s\n",ctime(&tStart)); 
	fprintf(fres,"MPI process rank: %d\n",rank); 
	
	fprintf(fres,"\nSEQUENCE FILE: %s\n",SEQUENCE_FILE); 
	fprintf(fres,"COUNT FILE   : %s\n",WEIS_COUNT_FILE); 
	
	fprintf(fres,"\nNumber of vectors   (GA_VECT)      %3d\n",GA_VECT); 
	fprintf(fres,"Number of eons      (GA_MAX_EON)   %3d\n",GA_MAX_EON); 
	fprintf(fres,"Survival rate       (GA_SURVIVE)    %6.3f\n",GA_SURVIVE); 
	fprintf(fres,"Mutation rate       (GA_MUTATE)     %6.3f\n",GA_MUTATE); 
	fprintf(fres,"Forecast interval   (FORECAST)     %3d\n",FORECAST); 
	
	fprintf(fres,"\nMODEL TYPE ");
#ifdef N_MODEL
	fprintf(fres,"N : ENTROPY WEIGHTS, -HIGH_WEIGHT ON FALSE NEGATIVES\n");
#else
#ifdef P_MODEL
	fprintf(fres,"P : ENTROPY WEIGHTS, -HIGH_WEIGHT ON FALSE POSITIVES\n");
#else
	fprintf(fres,"U : UNWEIGHTED MODEL\n");
#endif
#endif

	fflush(fres);				// *really* do the output, in case we kill the program 
} // Init_File

void File_Results(int nex)
// writes the results to fres at the end of evolution
{
int ka;
																									// write the aHistory array
	fprintf(fres,"\n===============\n\nExperiment %d\n\n",nex);
	for (ka=0; ka<GA_MAX_EON; ka++)
	if ((!ka) || (aHistory[ka-1].top != aHistory[ka].top))
		fprintf(fres,"%3d  %3d   %3d  %10.2f\n",ka,aHistory[ka].top, aHistory[ka].age,aHistory[ka].val - FITNESS_BASE); 
	fputc('\n',fres);

	fprintf(fres,"\nHIGH CONFLICT MODEL\n");
	Init_Saved_HMMGA(aHistory[GA_MAX_EON-1].top);  // move best high aFinal vector into aStates
	File_Model();
	
	fprintf(fres,"\nLOW CONFLICT MODEL\n");
	Init_Saved_HMMGA(aHistory[GA_MAX_EON-1].top + GA_VECT);  // move best low aFinal vector into aStates
	File_Model();

}	// File_Results

void File_Model(void)
// write the model currently in aState to fres
{  
int s, k;

fprintf(fres,"states: %d\n",max_states);			// write the vector 
  fprintf(fres,"symbols:  %d\n",max_symbols);
  for ( s=0; s < max_states; s++) 
  {
    fprintf(fres,"%8.4f\t",set_prev_trans(&aStates[s],-1.0));
    fprintf(fres,"%8.4f\t",set_recur_trans(&aStates[s],-1.0));
    fprintf(fres,"%8.4f\t",set_next_trans(&aStates[s],-1.0));
    for ( k=0; k < max_symbols; k++)
      fprintf(fres,"%8.4f\t",set_symb_out(&aStates[s],k,-1.0));
    fprintf(fres,"\n\n");
  }
  fprintf(fres,"\n");
} // File_Model

//__________________________________________________________________________________
//																																Balkans estimation

void Read_Count(void)
// read the Count file -- number of WEIS 22 events by week
// This routine also sets the various by-year indices
{
FILE 	*fin;
int 	iyear = 0;

	fin = fopen(WEIS_COUNT_FILE,"r");
	if (!fin) { printf("Error: can't open %s",WEIS_COUNT_FILE); exit(0); }
	
	printf("Reading %s\n",WEIS_COUNT_FILE);
	while (!feof(fin)) 
	{
		fscanf(fin, "%ld %d", &aCount22[N_Weeks].date, &aCount22[N_Weeks].count);
		if (aCount22[N_Weeks].date > YearStart[iyear])
		{
			Count_Year[iyear++] = N_Weeks;
			assert(iyear<N_YEARS);
			N_High[iyear]= N_High[iyear-1];
			N_Low[iyear] = N_Low[iyear-1];
		}

		if (aCount22[N_Weeks].count >= CONFL_THRESHOLD)
			N_High[iyear]++;
		else
			N_Low[iyear]++;
		N_Weeks++;
	}
	assert(N_Weeks<=MAX_WEEKS);
	N_Weeks--;
	Count_Year[iyear] = N_Weeks;
	
} // Read_Count


// Note: these procedures should be adjusting for the SEQ_LENGTH and FORECAST
//       in order to avoid getting templates that are too early in the sequence

long Rand_HighConflict(void)
// returns a randomly-chosen high conflict week
{
int ka = 0;
int kt = 0;
int	krand = (int)(UnifRnd() * N_High[Year]) + 1;  // randomly select number between 1 and N_High

	if (krand > N_High[Year]) krand =  N_High[Year]; // deal with the UnifRnd = 1 problem
	
	while (ka<Count_Year[Year])
	{
		if (aCount22[ka++].count >= CONFL_THRESHOLD) kt++;  // count high conflict weeks until we reach that number
		if (kt == krand) return (aCount22[ka-1].date);
	}
	assert(ka<Count_Year[Year]);		// kill program if we get here;
	return (aCount22[ka-1].date); 	// this should never occur
} // Rand_HighConflict

long Rand_LowConflict(void)
// returns a randomly-chosen low conflict week
// (this is essentially the same as Rand_HighConflict; see comments in that routine)
{
int ka = 0;
int kt = 0;
int	krand = (int)(UnifRnd() * N_Low[Year]) + 1;  // randomly select number between 1 and N_Low

	if (krand > N_Low[Year]) krand =  N_Low[Year];

	while (ka<Count_Year[Year])
	{
		if (aCount22[ka++].count < CONFL_THRESHOLD) kt++;
		if (kt == krand) return (aCount22[ka-1].date);
	}
	assert(ka<Count_Year[Year]);
	return (aCount22[ka-1].date);
} // Rand_LowConflict


void Get_Template_Dates(void)
//
{
#define MAX_TRY  4*MAX_TEMPLATES		// bail out if we exceed this

int 	ka, kb;
long	date;
int		ntry = 0;   // use to avoid infinite attempts to find templates

	aHigh[0] = Rand_HighConflict();
	for (ka=1;ka<MAX_TEMPLATES; ka++)
	{
		do
			date=Rand_HighConflict();
		while (Verify_Sequence_Start(date - FORECAST) < 0);		// check that we can get a sequence from this
		
		for (kb=0;kb<ka; kb++) if (date == aHigh[kb]) break;
		if (kb<ka) ka--;		// duplicate; try again
		else aHigh[ka] = date;
		if (++ntry > MAX_TRY) break;
	}
	
	if (ntry > MAX_TRY) {   // then the templates aren't there...
		fNoTemplates = TRUE;
		return;
	}

	ntry = 0;
	aLow[0] = Rand_LowConflict();
	for (ka=1;ka<MAX_TEMPLATES; ka++)
	{
		do
			date=Rand_LowConflict();
		while (Verify_Sequence_Start(date - FORECAST) < 0);
		
		for (kb=0;kb<ka; kb++) if (date == aLow[kb]) break;
		if (kb<ka) ka--;		// duplicate; try again
		else aLow[ka] = date;
		if (++ntry > MAX_TRY) break;
	}

	if (ntry > MAX_TRY) fNoTemplates = TRUE;

} // Get_Dates

int Verify_Sequence_Start(long ijul)
// Finds the aWEIS index of the beginning of a SEQ_SIZE sequence that ends on ijul
// If sequence cannot be created, returns -1
// Note that this is currently used only to confirm that a sequence exists
{
int ka=SEQ_SIZE-1;

		if (aWEIS_Date[ka] > ijul) return (-1);
		while (( aWEIS_Date[++ka] < ijul) && (ka < NumWEIS)) ;  // find the date in the array
		if (ka < NumWEIS)	return (ka - SEQ_SIZE + 1);
		else							return (-1);
		
} // Verify_Sequence_Start

void Init_Template_HMM(long *aDates)
// Reads a set of dates from aDates minus FORECAST; moves SEQ_SIZE events into strings, 
// then estimates a model
{
int		ka = 0, kb;
int		lastDate;
long	ijul;
int		kstring=0;

	for (kstring=0; kstring<MAX_TEMPLATES; kstring++)
	{		
		ijul = aDates[kstring] - FORECAST;
		lastDate=SEQ_SIZE-1;
		while (( aWEIS_Date[++lastDate] < ijul) && (lastDate < NumWEIS)) ;  // find the date in the array

		kb=0;
		for (ka = lastDate - SEQ_SIZE; ka < lastDate; ka++)			// load previous SEQ_SIZE events into strings
			strings[kstring][kb++] = aWEIS_Evt[ka];
		string_len[kstring] = SEQ_SIZE;
	}
	num_strings = MAX_TEMPLATES;
	
	// save a copy of the zero string, which will get overwritten in Compute_Running_Prob
	for (ka=0; ka<SEQ_SIZE; ka++) string_zero[ka] = strings[0][ka];

} // Init_Template_HMM

void Compute_Weekly_Prob (float *probs )
// computes alpha for SEQ_SIZE events prior to the beginning of a week, minus
// FORECAST days, based on aState model
{
int 		ka, kb, ijul,kweek; 
int			lastDate=0;
int 		firstwk,lastwk;			// time boundaries for calculations

// set globals
  trellis_width = SEQ_SIZE+1;  // define trellis dimensions
  max_symbols = MAX_SYMBOLS;
  string_len[0] = SEQ_SIZE;
	
	if (fFullForecast)
	{
		firstwk = 0;
		lastwk = N_Weeks;
	}
	else
	{
		firstwk = 0;
		lastwk = Count_Year[Year];
	}
	 	
	for (kweek=firstwk; kweek<lastwk; kweek++)
	{
		ijul = aCount22[kweek].date - FORECAST;

		while (( aWEIS_Date[++lastDate] < ijul) && (lastDate < NumWEIS)) {};  // find the date in the array

		if (lastDate - SEQ_SIZE < 0)			// too early to estimate
		{	probs[kweek] = 0.0;
			continue;}

		kb=0;
		for (ka = lastDate - SEQ_SIZE; ka < lastDate; ka++)			// load previous MAX_SEQ events into strings
			strings[0][kb++] = aWEIS_Evt[ka];											// Note that this is overwriting the first template

    probs[kweek] = test(strings[0],SEQ_SIZE);												// compute alpha
	}	

	// restore the zero string
	for (ka=0; ka<SEQ_SIZE; ka++) strings[0][ka] = string_zero[ka];

} // Compute_Weekly_Prob

void Write_Forecast (int isFull)
// write various summary results to fres.
// CAUTION: There is a heavy-duty side effect here -- Write_Forecast(TRUE) must be
//          called just *before*  Write_Forecast(FALSE) because the second call 
//          DOES NOT recompute aHighProb and aLowProb
{
int 		ka;
char		sres[7];			// tag for type of error
int			icorr = 0;		// # correct
int			itp = 0;			// # true positives
int			itn = 0;			// # true negatives
int			ifp = 0;			// # false positives
int			ifn = 0;			// # false negatives
int			ihigh = 0;		// # high conflict weeks
int			nweek = 0;		// # weeks where forecast was made
float		diff; 
int 		firstwk,lastwk;			// time boundaries for calculations

	printf("Writing results...\n");
	
	fprintf(fres,"\n==============================================\n");
	fprintf(fres,"\nTemplate dates:\n High    Low\n");
	for (ka=0;ka<MAX_TEMPLATES;ka++)
		fprintf(fres,"%6d   %6d\n",aHigh[ka],aLow[ka]);
		
																					
	if (isFull)
	{
		firstwk = 0;
		lastwk = N_Weeks;
		fprintf(fres,"\n\nAll Weeks\n");
		fFullForecast = TRUE;														// need to estimate future cases	 	
		Estimate_Vector(aHistory[GA_MAX_EON-1].top);		// load aHighProb and aLowProb using best vector
		fFullForecast = FALSE;													// [note this is using a side-effect of the call]
	}
	else
	{
		firstwk = Count_Year[Year];
		lastwk = N_Weeks;
		fprintf(fres,"\n\nPredicted Weeks Only\n");
	}

	for (ka=firstwk; ka<lastwk; ka++)
	{
		if (aHighProb[ka] == 0.0) continue;
		nweek++;
		if (aCount22[ka].count > CONFL_THRESHOLD) ihigh++;
		diff = aHighProb[ka] - aLowProb[ka];
		if ((aCount22[ka].count - CONFL_THRESHOLD) * diff >0) 
		{	icorr++;
			if (diff>0) 
			{ itp++;
				strcpy(sres,"OK+");}
			else
			{ itn++;
			  strcpy(sres,"OK-");}
		}
		else if (diff > 0)
		{	ifp++;
			strcpy(sres,"  FP");}			
		else
		{	ifn++;
			strcpy(sres,"    FN");}
			
		fprintf(fres,"%6d\t%3d\t%10.2f\t%10.3f\t%10.3f\t%s\n",aCount22[ka].date, aCount22[ka].count,aHighProb[ka],aLowProb[ka],diff,sres);
	}
	fprintf(fres,"\n%%High = %6.3f\n\n",(float)ihigh/(float)nweek);
	fprintf(fres,"\n%%OK   = %6.3f\n",(float)icorr/(float)nweek);
	fprintf(fres,"\n%%TP   = %6.3f\n",(float)itp/(float)nweek);
	fprintf(fres,"%%TN   = %6.3f\n",(float)itn/(float)nweek);
	fprintf(fres,"\n%%FP   = %6.3f\n",(float)ifp/(float)nweek);
	fprintf(fres,"%%FN   = %6.3f\n",(float)ifn/(float)nweek);
	
	fflush(fres);				// *really* do the output, in case we kill the program 
	
} // Write_Forecast

float Compare_High_Low(void)
// calculates comparison statistic for aHighProb[ka] - aLowProb[ka]
{
int 		ka;
int			icorr = 0;		// # correct
int			ifp = 0;			// # false positives
int			ifn = 0;			// # false negatives
int			nweek = 0;		// # weeks where forecast was made
float		sum = 0.0;
float		diff; 

#define HIGH_WEIGHT		1.68		// ln(p) weights
#define LOW_WEIGHT    0.21

	for (ka=0;ka<N_Weeks;ka++)
	{
		if (aHighProb[ka] == 0.0) continue;
		nweek++;
		diff = aHighProb[ka] - aLowProb[ka];
		if ((aCount22[ka].count - CONFL_THRESHOLD) * diff >0) 
		{	icorr++;
			if (diff>0) sum += HIGH_WEIGHT;
			else  			sum += LOW_WEIGHT;}
		else if (diff > 0)
		{	ifp++;
#ifdef P_MODEL
			sum -= HIGH_WEIGHT;			
#else			
			sum -= LOW_WEIGHT;
#endif
		}
		else
		{	ifn++;
#ifdef P_MODEL
			sum -= LOW_WEIGHT;
#else			
			sum -= HIGH_WEIGHT;			
#endif
		}
	}
	return (((FITNESS_BASE + sum) > 0) ? (FITNESS_BASE + sum) : 0.25);  // use the 0.25 to prevent DIV 0 problems
	
} // Compare_High_Low

